/**************************************************************************************************

Main.C -- (C) Niklas Een, 2003

Read a DIMACS file and apply the SAT-solver to it.

**************************************************************************************************/


#include "Solver.h"
#include <ctime>


//=================================================================================================
// Helpers:


// Reads an input stream to end-of-file and returns the result as a 'char*' terminated by '\0'.
//
char* readFile(FILE* in, int& out_size = *(int*)NULL)
{
    char*   data = xmalloc<char>(65536);
    int     cap  = 65536;
    int     size = 0;

    while (!feof(in)){
        if (size == cap){
            cap *= 2;
            data = xrealloc(data, cap); }
        size += fread(&data[size], 1, 65536, in);
    }
    data = xrealloc(data, size+1);
    data[size] = '\0';

    if (&out_size != NULL)
        out_size = size;

    return data;
}


//=================================================================================================
// DIMACS Parser:


static void skipWhitespace(char*& in) {
    while ((*in >= 9 && *in <= 13) || *in == 32)
        in++; }

static void skipLine(char*& in) {
    for (;;){
        if (*in == 0) return;
        if (*in == '\n') { in++; return; }
        in++; } }

static int parseInt(char*& in) {
    int     val = 0;
    bool    neg = false;
    skipWhitespace(in);
    if      (*in == '-') neg = true, in++;
    else if (*in == '+') in++;
    if (*in < '0' || *in > '9') fprintf(stderr, "PARSE ERROR! Unexpected char: %c\n", *in), exit(1);
    while (*in >= '0' && *in <= '9')
        val = val*10 + (*in - '0'),
        in++;
    return neg ? -val : val; }

static void readClause(char*& in, Solver& S, vec<Lit>& lits) {
    int     parsed_lit, var;
    lits.clear();
    for (;;){
        parsed_lit = parseInt(in);
        if (parsed_lit == 0) break;
        var = abs(parsed_lit)-1;
        while (var >= S.nVars()) S.newVar();
        lits.push( (parsed_lit > 0) ? Lit(var) : ~Lit(var) );
    }
}

static bool parse_DIMACS_main(char* in, Solver& S) {
    vec<Lit>    lits;
    for (;;){
        skipWhitespace(in);
        if (*in == 0)
            break;
        else if (*in == 'c' || *in == 'p')
            skipLine(in);
        else{
            readClause(in, S, lits);
            if (!S.addClause(lits))
                return false;
        }
    }
    return S.simplifyDB();
}


// Inserts problem into solver. Returns FALSE upon immediate conflict.
//
bool parse_DIMACS(FILE* in, Solver& S) {
    char* text = readFile(in);
    bool ret = parse_DIMACS_main(text, S);
    free(text);
    return ret; }


//=================================================================================================


void printStats(SolverStats& stats, double cpu_time)
{
    printf("starts                : %lld\n", stats.starts);
    printf("conflicts             : %lld\n", stats.conflicts);
    printf("decisions             : %lld\n", stats.decisions);
    printf("propagations          : %lld\n", stats.propagations);
    printf("inspects              : %lld\n", stats.inspects);
    printf("CPU time              : %g s\n", cpu_time);
}


//=================================================================================================


int main(int argc, char** argv)
{
    Solver      S;
    bool        st;

    if (argc == 1)
        st = parse_DIMACS(stdin, S);
    else{
        FILE* in = fopen(argv[1], "rb");
        if (in == NULL)
            fprintf(stderr, "ERROR! Could not open file: %s\n", argv[1]),
            exit(1);
        st = parse_DIMACS(in, S);
        fclose(in);
    }
    if (!st)
        printf("Trivial problem\nUNSATISFIABLE\n"),
        exit(20);

    clock_t T0 = clock();
    S.verbosity = 1;
    st = S.solve();
    clock_t T1 = clock();
    printStats(S.stats, (double)(T1-T0) / CLOCKS_PER_SEC);
    printf("\n");
    printf(st ? "SATISFIABLE\n" : "UNSATISFIABLE\n");

    return 0;
}
