/**************************************************************************************************

Solver.h -- (C) Niklas Een, 2003

A simple Chaff-like SAT-solver with support for incremental SAT.

**************************************************************************************************/

#ifndef Solver_h
#define Solver_h

#include "SolverTypes.h"
#include "Constraints.h"
#include "Queue.h"
#include "VarOrder.h"


//=================================================================================================
// Solver -- the main class:


struct SolverStats {
    int64   starts, decisions, propagations, inspects, conflicts;
    SolverStats(void) : starts(0), decisions(0), propagations(0), inspects(0), conflicts(0) { }
};


struct SearchParams {
    double  var_decay, clause_decay;    // (reasonable values are: 0.95, 0.999, 0.02)    
    SearchParams(double v = 1, double c = 1) : var_decay(v), clause_decay(c) { }
};


class Solver {
protected:
    vec<Constr*>        constrs;        // List of problem constraints.
    vec<Clause*>        learnts;        // List of learnt clauses.
    double              cla_inc;        // Amount to bump next clause with.
    double              cla_decay;      // INVERSE decay factor for clause activity: stores 1/decay.

    vec<double>         activity;       // A heuristic measurement of the activity of a variable.
    double              var_inc;        // Amount to bump next variable with.
    double              var_decay;      // INVERSE decay factor for variable activity: stores 1/decay. Use negative value for static variable order.
    VarOrder            order;          // Keeps track of the decision variable order.

    vec<vec<Constr*> >  watches;        // 'watches[lit]' is a list of constraints watching 'lit' (will go there if literal becomes true).
    vec<vec<Constr*> >  undos;          // 'undos[var]' is a list of constraints that will be called when 'var' becomes unbound.
    Queue<Lit>          propQ;          // Propagation queue.

    vec<char>           assigns;        // The current assignments (lbool:s stored as char:s).
    vec<Lit>            trail;          // List of assignments made. 
    vec<int>            trail_lim;      // Separator indices for different decision levels in 'trail'.
    vec<Constr*>        reason;         // 'reason[var]' is the clause that implied the variables current value, or 'NULL' if none.
    vec<int>            level;          // 'level[var]' is the decision level at which assignment was made.
    int                 root_level;     // Level of first proper decision.

    double              random_seed;    // For the internal random number generator (makes solver deterministic over different platforms).

    // Main internal methods:
    //
    bool        assume     (Lit p);
    void        undoOne    (void);
    void        cancel     (void);
    void        cancelUntil(int level);
    void        record     (const vec<Lit>& clause);

    void        analyze    (Constr* confl, vec<Lit>& out_learnt, int& out_btlevel); // (bt = backtrack)
    bool        enqueue    (Lit fact, Constr* from = NULL);
    Constr*     propagate  (void);
    void        reduceDB   (void);
    lbool       search     (int nof_conflicts, int nof_learnts, const SearchParams& params);
    double      progressEstimate(void);

    // Activity:
    //
    void    varBumpActivity(Lit p) {
        if (var_decay < 0) return;     // (negative decay means static variable order -- don't bump)
        if ( (activity[var(p)] += var_inc) > 1e100 ) varRescaleActivity();
        order.update(var(p)); }
    void    varDecayActivity(void) { if (var_decay >= 0) var_inc *= var_decay; }
    void    varRescaleActivity(void);

    void    claBumpActivity(Clause* c) { if ( (c->activity() += cla_inc) > 1e20 ) claRescaleActivity(); }
    void    claDecayActivity(void) { cla_inc *= cla_decay; }
    void    claRescaleActivity(void);

    // Helpers:
    int     nAssigns(void) { return trail.size(); }
    int     nConstrs(void) { return constrs.size(); }
    int     nLearnts(void) { return learnts.size(); }

    lbool   value(Var x) { return toLbool(assigns[x]); }
    lbool   value(Lit p) { return sign(p) ? ~toLbool(assigns[var(p)]) : toLbool(assigns[var(p)]); }

    int     decisionLevel(void) { return trail_lim.size(); }

public:
    Solver(void) : cla_inc          (1)
                 , cla_decay        (1)
                 , var_inc          (1)
                 , var_decay        (1)
                 , order            (assigns, activity)
                 , random_seed      (91648253)
                 , progress_estimate(0)
                 , verbosity(0)
                 { }
   ~Solver(void) {
        for (int i = 0; i < learnts.size(); i++) learnts[i]->remove(*this, true);
        for (int i = 0; i < constrs.size(); i++) constrs[i]->remove(*this, true); }


    // Statistics: (read-only member variable)
    //
    SolverStats stats;


    // Problem specification:
    //
    Var     newVar(void);
    int     nVars(void)        { return assigns.size(); }

    // -- constraints:
    friend class Clause;
    friend class AtMost;
    friend bool Clause_new(Solver& S, const vec<Lit>& ps, bool learnt, Clause*& out_clause);
    friend bool AtMost_new(Solver& S, const vec<Lit>& ps, int  max   , AtMost*& out_constr);

    bool    addClause(const vec<Lit>& ps)        { Clause* c; bool ret = Clause_new(*this, ps, false, c); if (c != NULL) constrs.push(c); return ret; }
    bool    addAtMost(const vec<Lit>& ps, int n) { AtMost* c; bool ret = AtMost_new(*this, ps, n    , c); if (c != NULL) constrs.push(c); return ret; }

    // Solving:
    //
    bool    simplifyDB(void);
    bool    solve(const vec<Lit>& assumps = vec<Lit>());

    double      progress_estimate;  // Set by 'search()'.
    vec<bool>   model;              // If problem is solved, this vector contains the model (if any).
    int         verbosity;          // Verbosity level. 0=silent, 1=some progress report, 2=everything
};


//=================================================================================================
#endif
