/**************************************************************************************************

VarOrder.C -- (C) Niklas Een, 2003

ADT for maintaining the variable ordering. It will keep a list of all variables sorted on the
current activity.  

**************************************************************************************************/


#ifndef VarOrder_h
#define VarOrder_h

#include "SolverTypes.h"


//=================================================================================================


class VarOrder {

    const vec<char>&    assigns;    // var->val. Pointer to external assignment table.
    const vec<double>&  activity;   // var->act. Pointer to external activity table.
    vec<int>            varpos;     // var->pos.
    vec<Var>            order;      // pos->var.  

    int     last_var;

public:
    VarOrder(const vec<char>& ass, const vec<double>& act) : assigns(ass), activity(act), last_var(0) { }

    inline void newVar(void);
    inline void update(Var x);                 // Called when variable increased in activity.
           void updateAll(void);               // Called when all variables have been assigned new activity.
    inline void undo(Var x);                   // Called when variable is unassigned and may be selected again.
    inline Var  select(void);                  // Selects a new, unassigned variable (or 'var_Undef' if none exists).
};


void VarOrder::newVar(void)
{
    varpos.push(varpos.size());
    order.push(order.size());
}


void VarOrder::update(Var x)
{
    for (int i = varpos[x]; i > 0 && activity[order[i-1]] < activity[x]; i--){
        assert(order[i] == x);
        varpos[order[i-1]]++; varpos[order[i]]--;
        order[i-1] ^= order[i]; order[i] ^= order[i-1]; order[i-1] ^= order[i];
    }
}


void VarOrder::undo(Var x)
{
    if (varpos[x] < last_var)
        last_var = varpos[x];
}


Var VarOrder::select(void)
{
    for (int i = last_var; i < order.size(); i++){
        if (toLbool(assigns[order[i]]) == l_Undef){
            last_var = i;
            return order[i]; }
    }
    return var_Undef;
}


//=================================================================================================
#endif
